import type { HypervalError, Trace } from './Trace';
interface Test<T> {
    /**
     * The Hyper test function
     *
     * @param raw Data to test.
     * @param failure The `Trace` object tracking current validation node.
     */
    hyper(raw: any, failure: Trace): T;
}
export declare type HyperConstructor<T> = [
    string,
    (raw: any, failure: Trace) => T
];
declare type WrapTest<T, U> = Test<U>;
/**
 * Base Hyper test class
 */
export declare class Hyper<T, U = T> implements WrapTest<T, U> {
    /** Message to print on failure */
    onfail: HyperConstructor<U>[0];
    /** Hyper test function */
    hyper: HyperConstructor<U>[1];
    /**
     * New Hyper Object
     *
     * @param onfail Message to print on failure.
     * @param hyper Hyper test function.
     */
    constructor(
    /** Message to print on failure */
    onfail: HyperConstructor<U>[0], 
    /** Hyper test function */
    hyper: HyperConstructor<U>[1]);
    /**
     * Checks a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw typed raw value.
     * @returns checked and altered Value
     */
    apply(raw: T, strict: 'strict'): U;
    /**
     * Checks a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw untyped Raw value.
     * @returns checked and altered Value
     */
    apply(raw: unknown): U;
    /**
     * Asserts a value.
     * Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, asserted data will be of input type.
     *
     * @param raw typed raw value.
     * @returns asserts is type
     */
    assert(raw: T, strict: 'strict'): asserts raw is T;
    /**
     * Asserts a value.
     * Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, asserted data will be of input type.
     *
     * @param raw untyped raw value.
     * @returns asserts is type
     */
    assert(raw: unknown): asserts raw is T;
    /**
     * Typeguard a value.
     * Alter and cast modifications are still applied to an unused copy in the background and discarded as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, typeguarded data will be of input type.
     *
     * @param raw typed raw value.
     * @returns is type
     */
    is(raw: T, strict: 'strict'): raw is T;
    /**
     * Typeguard a value.
     * Alter and cast modifications are still applied to an unused copy in the background and discarded as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, typeguarded data will be of input type.
     *
     * @param raw untyped raw value.
     * @returns is type
     */
    is(raw: unknown): raw is T;
    /**
     * Validates a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw typed raw value.
     * @returns Raw validation object
     */
    validate(raw: T, strict: 'strict'): {
        error: HypervalError;
        result?: undefined;
    } | {
        result: U;
        error?: undefined;
    };
    /**
     * Validates a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw untyped raw value.
     * @returns Raw validation object
     */
    validate(raw: unknown): {
        error: HypervalError;
        result?: undefined;
    } | {
        result: U;
        error?: undefined;
    };
    /**
     * Checks a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * Use this method when validating against promises.
     *
     * @param raw typed raw value.
     * @returns checked and altered Value
     */
    thenApply(raw: Promise<T> | T, strict: 'strict'): Promise<U>;
    /**
     * Checks a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * Use this method when validating against promises.
     *
     * @param raw untyped raw value.
     * @returns checked and altered Value
     */
    thenApply(raw: Promise<unknown> | unknown): Promise<U>;
    /**
     * Asserts a value.
     * Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, asserted data will be of input type.
     *
     * Use this method when validating against promises.
     *
     * @param raw typed raw value.
     * @returns void
     */
    thenAssert(raw: Promise<T> | T, strict: 'strict'): Promise<void>;
    /**
     * Asserts a value.
     * Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, asserted data will be of input type.
     *
     * Use this method when validating against promises.
     *
     * @param raw untyped raw value.
     * @returns void
     */
    thenAssert(raw: Promise<unknown> | unknown): Promise<void>;
    /**
     * Typeguard a value.
     * Alter and cast modifications are still applied to an unused copy in the background and discarded as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, typeguarded data will be of input type.
     *
     * Use this method when validating against promises.
     *
     * @param raw typed raw value.
     * @returns is type
     */
    thenIs(raw: Promise<T> | T, strict: 'strict'): Promise<boolean>;
    /**
     * Typeguard a value.
     * Alter and cast modifications are still applied to an unused copy in the background and discarded as part of this process.
     * Original value remains unaltered.
     * If used with data that has cast and alter modifications, typeguarded data will be of input type.
     *
     * Use this method when validating against promises.
     *
     * @param raw untyped raw value.
     * @returns is type
     */
    thenIs(raw: Promise<unknown> | unknown): Promise<boolean>;
    /**
     * Validates a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * Use this method when validating against promises.
     *
     * @param raw typed raw value.
     * @returns Raw validation object
     */
    thenValidate(raw: Promise<T> | T, strict: 'strict'): Promise<{
        error: HypervalError;
        result?: undefined;
    } | {
        result: U;
        error?: undefined;
    }>;
    /**
     * Validates a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * Use this method when validating against promises.
     *
     * @param raw untyped raw value.
     * @returns Raw validation object
     */
    thenValidate(raw: Promise<unknown> | unknown): Promise<{
        error: HypervalError;
        result?: undefined;
    } | {
        result: U;
        error?: undefined;
    }>;
    /**
     * @deprecated Use {@link apply}
     * Checks a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw Raw value.
     * @returns Checked Value
     */
    anyApply(raw: any): U;
    /**
     * @deprecated Use {@link assert}
     * Typeguard a value. Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered. If used with data that has cast and alter modifications, asserted data may be incorrect.
     *
     * @param raw Raw value.
     * @returns Checked Value
     */
    anyAssert(raw: any): void;
    /**
     * @deprecated Use {@link is}
     * Typeguard a value. Alter and cast modifications are still applied to an unused copy in the background as part of this process.
     * Original value remains unaltered. If used with data that has cast and alter modifications, asserted data may be incorrect.
     *
     * @param raw Raw value.
     * @returns Checked Value
     */
    anyIs(raw: any): raw is T;
    /**
     * @deprecated Use {@link validate}
     * Validates a value, applying any alter and cast modifications.
     * Returns a copy with altered and casted data.
     * Does not mutate original but may refer back to it.
     *
     * @param raw Raw value.
     * @returns Checked Value
     */
    anyValidate(raw: any): {
        error: HypervalError;
        result?: undefined;
    } | {
        result: U;
        error?: undefined;
    };
    /**
     * New Hyper Object
     *
     * @param onfail Message to print on failure.
     * @param hyper Hyper test function.
     */
    static new<T, U = T>(onfail: HyperConstructor<U>[0], test: HyperConstructor<U>[1]): Hyper<T, U>;
}
/**
 * Generates a generic test function.
 *
 * @param name Name of test.
 * @param test Test function that validates input.
 */
export declare function hyperval_factory<T>(name: string, test: (x: any) => boolean): Hyper<T>;
/** @deprecated Use {@link HyperRaw} */
export declare type Hyperval<T> = HyperRaw<T>;
/** @deprecated Use {@link HyperVal} */
export declare type Hyperout<T> = HyperVal<T>;
/**
 * Get the input value of a hyper if there is no mutations, otherwise get the output value.
 */
export declare type HyperVal<T> = T extends Hyper<infer A, infer B> ? (B extends A ? (A extends B ? A : B) : B) : T;
/**
 * Get the input value of a hyper
 */
export declare type HyperRaw<T> = T extends Hyper<infer U, any> ? U : T;
export {};
