const START = Symbol('START');
/**
 * The failure error used internally to track error types and status.
 */
export class HypervalError extends Error {
    /**
     * Create new HypervalError
     *
     * @param source Source data where the error occurred.
     * @param path Path on source where the error occurred.
     * @param validation The name of the validation error.
     */
    constructor(source, path, validation) {
        super(`Hyperval Validation Error: <${validation}>`);
        this.source = source;
        this.path = path;
        this.validation = validation;
        this.name = `Hyperval Validation Error: <${validation}>`;
        this.message = this.full;
    }
    /**
     * Returns the full error result. Suitable for printing and debugging.
     */
    get full() {
        let o = this.source;
        for (const p of this.path) {
            o = o === null || o === void 0 ? void 0 : o[p];
        }
        return `${this.name}:

[${this.path.join(' -> ')}]: ${JSON.stringify(o, undefined, 2)}`;
    }
}
/**
 * Trace where validation errors occur.
 */
export class Trace {
    constructor(root, prev = START, current = START) {
        this.root = root;
        this.prev = prev;
        this.current = current;
        this.fork = this.fork.bind(this);
        this.throw = this.throw.bind(this);
    }
    /**
     * Split the trace path. Used when going down an object or array level.
     *
     * @param name Name of forked path.
     * @returns `Trace` New copy of trace for the forked path.
     */
    fork(name) {
        return new Trace(this.root, this, name);
    }
    /**
     * Throws a HypervalError error at the current path of the trace.
     *
     * @param error Error to throw.
     */
    throw(error) {
        const error_path = this.current === START ? [] : [this.current];
        let parent = this.prev;
        let lim = 10;
        while (parent !== START && parent.current !== START && --lim > 0) {
            error_path.push(parent.current);
            parent = parent.prev;
        }
        throw new HypervalError(this.root, error_path.reverse(), error);
    }
}
